// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#ifndef _OUTPUTRESULTSINK_H_
#define _OUTPUTRESULTSINK_H_

#include <istream>
#include <ostream>
#include <limits>
#include <queue>
#include <string>
#include <pthread.h>

// Choice of how to write to output file.
enum OutputResultSinkMode {
	ORSM_SYNC,
	ORSM_ASYNC_FIFO,
	ORSM_ASYNC_INORDER
};

enum MappingResultType {
	MRT_MAPPED,
	// Only one of the reads of a pair must be mapped.
	MRT_HALFMAPPED,
	MRT_UNMAPPED,
	MRT_INVALID
};

class MappingResult {
public:
	unsigned readNum;
	std::istream* stream;
	MappingResultType type;

	MappingResult(unsigned readNum, std::istream* stream, MappingResultType type) :
		readNum(readNum), stream(stream), type(type) {
	}

	MappingResult() :
		readNum(std::numeric_limits<unsigned>::max()), stream(NULL), type(MRT_INVALID) {
	}

	void clear() {
		if (stream) {
			delete stream;
			stream = NULL;
		}
		type = MRT_INVALID;
		readNum = std::numeric_limits<unsigned>::max();
	}

	bool isValid() {
		return (type != MRT_INVALID && stream != NULL);
	}

	// For ordering within a heap.
	bool operator<(const MappingResult& mr) const {
		return readNum >= mr.readNum;
	}
};

class OutputResultSink {
protected:
	std::ostream& mappedOut;
	void writeResult(MappingResult& result);
	unsigned resultsWritten;

public:
	OutputResultSink(std::ostream& mappedOut) :
		mappedOut(mappedOut) {
	}

	// To be implemented by sub-class.
	// Protocol: caller is responsible for allocating result.stream
	// but result.stream is deleted within this method.
	virtual void add(MappingResult& result) = 0;

	void writeInitialString(std::istream* initialString) {
		if (initialString->rdbuf()->in_avail() > 0) {
			mappedOut << initialString->rdbuf();
		}
	}

	unsigned getResultsWritten() {
		return resultsWritten;
	}
};

class SyncOutputResultSink : public OutputResultSink {
protected:
	pthread_mutex_t outputMutex;

public:
	SyncOutputResultSink(std::ostream& mappedOut) :
			OutputResultSink(mappedOut) {
		pthread_mutex_init(&outputMutex, 0);
	}

	void add(MappingResult& result);

	virtual ~SyncOutputResultSink() {
		pthread_mutex_destroy(&outputMutex);
	}
};

class AsyncOutputResultSink : public OutputResultSink {
protected:
	pthread_mutex_t queueMutex;
	pthread_mutex_t terminateMutex;
	pthread_cond_t flushCond;
	pthread_cond_t stallCond;
	unsigned maxOutstanding;
	bool done;

	// To be implemented by subclasses to provide different ordering guarantees.
	virtual void push(MappingResult& result) = 0;
	virtual void pop(MappingResult& result) = 0;

	// To be implemented by subclasses.
	virtual bool sleep() const = 0;
	virtual bool wakeup() const = 0;

public:
	AsyncOutputResultSink(std::ostream& mappedOut, unsigned maxOutstanding) :
			OutputResultSink(mappedOut), maxOutstanding(maxOutstanding), done(false) {
		pthread_mutex_init(&queueMutex, NULL);
		pthread_mutex_init(&terminateMutex, NULL);
		pthread_cond_init(&flushCond, NULL);
		pthread_cond_init(&stallCond, NULL);
	}

	virtual void add(MappingResult& result);

	// Meant to be run as a thread.
	void flushResults();

	// Cleanly terminate thread.
	void terminate();

	virtual ~AsyncOutputResultSink() {
		pthread_mutex_destroy(&queueMutex);
		pthread_mutex_destroy(&terminateMutex);
		pthread_cond_destroy(&flushCond);
		pthread_cond_destroy(&stallCond);
	}

	virtual unsigned getQueueSize() const = 0;
};

class AsyncFifoOutputResultSink : public AsyncOutputResultSink {
protected:
	// FIFO queue.
	std::queue<MappingResult> resultQueue;

	void push(MappingResult& result);
	void pop(MappingResult& result);

	bool sleep() const {
		return resultQueue.empty() && !done;
	}

	bool wakeup() const {
		return true;
	}

public:
	AsyncFifoOutputResultSink(std::ostream& mappedOut, unsigned maxOutstanding) :
		AsyncOutputResultSink(mappedOut, maxOutstanding) {
	}

	unsigned getQueueSize() const {
		return resultQueue.size();
	}
};

class AsyncInOrderOutputResultSink : public AsyncOutputResultSink {
protected:
	// Used as basis for a min heap.
	std::vector<MappingResult> resultQueue;
	// Assuming actual rids start from 1 and not 0.
	unsigned lastRid;

	void push(MappingResult& result);
	void pop(MappingResult& result);

	bool sleep() const;
	bool wakeup() const;

public:
	AsyncInOrderOutputResultSink(std::ostream& mappedOut, unsigned maxOutstanding) :
			AsyncOutputResultSink(mappedOut, maxOutstanding), lastRid(0) {
		make_heap(resultQueue.begin(), resultQueue.end());
	}

	unsigned getQueueSize() const {
		return resultQueue.size();
	}
};

#endif
