// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#ifndef _SINGLEENDSEARCHER_H_
#define _SINGLEENDSEARCHER_H_

#include "search/searcher.h"
#include "search/singleendverifier.h"
#include "search/singleendmerger.h"

#include <vector>

class SingleEndSearcher : public Searcher {
public:
	std::vector<unsigned> forwardCandidates;
	std::vector<ReadMapping> forwardResults;
	std::vector<unsigned> reverseCandidates;
	std::vector<ReadMapping> reverseResults;

	Read forwardRead;
	Read reverseRead;

	SingleEndSearcher(Indexer* indexer, unsigned distThresh, SingleEndMerger* merger,
			SingleEndVerifier* verifier, HobbesConf& conf, HobbesStats& stats) :
		Searcher(indexer, distThresh, conf, stats), merger(merger), verifier(verifier) {
	}

	MappingResultType getMappingResultType() {
		if (forwardResults.size() == 0 && reverseResults.size() == 0) {
			return MRT_UNMAPPED;
		} else {
			return MRT_MAPPED;
		}
	}

	virtual ~SingleEndSearcher() {
		delete merger;
	}

protected:
	SingleEndMerger* merger;
	SingleEndVerifier* verifier;

	void prepare(unsigned readLen);
	void findAllMappings(ReadInfo& readInfo);
	void findKMappings(ReadInfo& readInfo);
	void getStats(std::pair<unsigned, unsigned>& counts);
	void getBestStats();

	void genReverses(ReadInfo& readInfo);
	void mergeAndVerifyAll(Read& read, std::vector<unsigned>& candidates,
			std::vector<ReadMapping>& results);
	bool mergeAndVerifyK(bool exactMatch, unsigned& k, Read& read,
			std::vector<unsigned>& candidates, std::vector<ReadMapping>& results);
};

class SingleEndHammingSearcher: public SingleEndSearcher {
public:
	SingleEndHammingSearcher(Indexer* indexer, unsigned distThresh, HobbesConf& conf,
			HobbesStats& stats) :
				SingleEndSearcher(indexer, distThresh,
						new SingleEndHammingMerger(indexer, distThresh),
						new SingleEndHammingVerifier(indexer, distThresh), conf, stats) {
		forwardCandidates.reserve(CANDIS_PREALLOC_SIZE);
		forwardResults.reserve(RESULTS_PREALLOC_SIZE);
		reverseCandidates.reserve(CANDIS_PREALLOC_SIZE);
		reverseResults.reserve(RESULTS_PREALLOC_SIZE);
	}
};

class SingleEndIndelSearcher: public SingleEndSearcher {
public:
	SingleEndIndelSearcher(Indexer* indexer, unsigned distThresh, HobbesConf& conf,
			HobbesStats& stats) :
				SingleEndSearcher(indexer, distThresh,
						new SingleEndIndelMerger(indexer, distThresh),
						new SingleEndIndelVerifier(indexer, distThresh), conf, stats) {
		forwardCandidates.reserve(CANDIS_PREALLOC_SIZE);
		forwardResults.reserve(RESULTS_PREALLOC_SIZE);
		reverseCandidates.reserve(CANDIS_PREALLOC_SIZE);
		reverseResults.reserve(RESULTS_PREALLOC_SIZE);
	}
};

class SingleEndHeuristicIndelSearcher: public SingleEndSearcher {
public:
	SingleEndHeuristicIndelSearcher(Indexer* indexer, unsigned distThresh, HobbesConf& conf,
			HobbesStats& stats) :
				SingleEndSearcher(indexer, distThresh,
						new SingleEndHeuristicIndelMerger(indexer, distThresh),
						new SingleEndHeuristicIndelVerifier(indexer, distThresh), conf, stats) {
		forwardCandidates.reserve(CANDIS_PREALLOC_SIZE);
		forwardResults.reserve(RESULTS_PREALLOC_SIZE);
		reverseCandidates.reserve(CANDIS_PREALLOC_SIZE);
		reverseResults.reserve(RESULTS_PREALLOC_SIZE);
	}
};

#endif
