// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#ifndef _SINGLEENDVERIFIER_H_
#define _SINGLEENDVERIFIER_H_

#include "search/verifier.h"

#include <sstream>
#include <string>
#include <vector>
#include <cstring>

#ifndef COUNT_FILTER_GRAM_LEN
#define COUNT_FILTER_GRAM_LEN 3
// pow(4, COUNT_FILTER_GRAM_LEN)
#define GRAM_COUNT_ARRAY_LEN (4 << (COUNT_FILTER_GRAM_LEN * 2))
#endif

#ifndef CACHE_PREFETCH_STEP
#define CACHE_PREFETCH_STEP 2
#endif

struct GramCountArray {
	unsigned count[GRAM_COUNT_ARRAY_LEN];
	GramCountArray() {
		memset(count, 0, GRAM_COUNT_ARRAY_LEN * sizeof(unsigned));
	}
};

class ReadMapping {
public:
	unsigned pos;
	unsigned distance;
	//const std::stringstream* cigar;
	std::string cigar;

	//ReadMapping(unsigned pos, unsigned distance, const std::stringstream* cigar) :
	//		pos(pos), distance(distance), cigar(cigar) { }
	ReadMapping(unsigned pos, unsigned distance, std::string cigar) :
			pos(pos), distance(distance), cigar(cigar) { }
/*
	virtual void clear() {
		if (cigar != NULL) {
			delete cigar;
		}
	}
*/

//	~ReadMapping() {
//		  if(cigar != NULL) {
//			 delete cigar;
//			 cigar = NULL;
//		  }
//	   }
};

class SingleEndVerifier: public Verifier {
protected:
    unsigned long t_D[6][6], t_U[6];
    unsigned long t_v, t_vp, t_m, t_mask;
    int t_g, t_k, t_i;
    int t_index;
    int t_max;

public:

	SingleEndVerifier(Indexer* indexer, unsigned distThresh) :
		Verifier(indexer, distThresh)
    {

	}

	virtual void verifyCandidates(const std::string& read, std::vector<unsigned>& candidates,
			std::vector<ReadMapping>& results, bool unique) = 0;

	virtual void verifyCandidates(unsigned& startPos, unsigned& k, const std::string& read,
			std::vector<unsigned>& candidates, std::vector<ReadMapping>& results) = 0;

	virtual unsigned getDistance(const char* candi, unsigned candiLen, const std::string& read,
			int& offset, unsigned& refend) = 0;

	virtual unsigned getDistance(const char* candi, unsigned candiLen,
			const std::string& read) = 0;

	void verifyCandidatesExact(unsigned& startPos, unsigned& k, const std::string& read,
			std::vector<unsigned>& candidates, std::vector<ReadMapping>& results);

	bool exactMatch(const char* candi, unsigned candiLen, const std::string& read);
};

class SingleEndHammingVerifier: public SingleEndVerifier {
public:
	SingleEndHammingVerifier(Indexer* indexer, unsigned distThresh) :
		SingleEndVerifier(indexer, distThresh) {
	}

	void verifyCandidates(const std::string& read, std::vector<unsigned>& candidates,
			std::vector<ReadMapping>& results, bool unique);

	void verifyCandidates(unsigned& startPos, unsigned& k, const std::string& read,
			std::vector<unsigned>& candidates, std::vector<ReadMapping>& results);

	unsigned getDistance(const char* candi, unsigned candiLen, const std::string& read,
			int& offset, unsigned& refend);

	unsigned getDistance(const char* candi, unsigned candiLen, const std::string& read);
};

class SingleEndIndelVerifier : public SingleEndVerifier {
public:
	char readCopy[CSTR_PREALLOC];
	char candiCopy[CSTR_PREALLOC];
	char scores1D0[400];
	char scores1D1[400];
//	char maxreflen[400];
	char * scores0;
	char * scores1;
	char * ptrs;


	SingleEndIndelVerifier(Indexer* indexer, unsigned distThresh) :
				SingleEndVerifier(indexer, distThresh), gramLen(indexer->gramGen.gramLen),
				matrix(new unsigned[ED_MATRIX_DIM + 1][ED_MATRIX_DIM + 1]),
				latestPosChecked(0) {
	}

	void verifyCandidates(const std::string& read, std::vector<unsigned>& candidates,
			std::vector<ReadMapping>& results, bool unique);

	void verifyCandidates(unsigned& startPos, unsigned& k, const std::string& read,
			std::vector<unsigned>& candidates, std::vector<ReadMapping>& results) {
	}

	unsigned getDistance(const char* candi, unsigned candiLen, const std::string& read);

	unsigned getDistance(const char* candi, unsigned candiLen, const std::string& read,
			int& offset, unsigned& refend);

	unsigned getCigar(const char* candi, const std::string& read, std::stringstream& cigar,
				unsigned refMatchLen, unsigned dist);

	void fillGramCountArray(const char* candi, unsigned candiLen, GramCountArray& counts);

	unsigned getDistanceLowerBound(GramCountArray& readCounts, unsigned refPos, unsigned refLen);

	unsigned getBitVectorEditDistance(unsigned candi, unsigned& refMatchLen, unsigned maxLen,
			const std::string& query, unsigned remainingDist,std::stringstream& cigar,
			unsigned& indel, int& offset);
	unsigned getBitVectorEditDistance(const char* candi, unsigned& refMatchLen, unsigned maxLen,
			const std::string& query, unsigned remainingDist);

	int TEditDistance(int k, char * s1, char * s2, int l1);
    int Thanh1_Align(int allowed_errors, char * s1, char * s2, int reflen, int redlen, int z);
    int BandedAlign(int allowed_errors, char * s1, char * s2, int reflen, int redlen, int z);
    int BandedAlign(int allowed_errors, char * s1, char * s2, int reflen, int redlen, int z,std::stringstream& cigar,
		unsigned& indel, int& offset);
    int Thanh_Align(int allowed_errors, char * s1, char * s2, int reflen, int redlen, int z);
	~SingleEndIndelVerifier() {
		delete[] matrix;
	}

protected:
	unsigned gramLen;
	unsigned (*matrix)[ED_MATRIX_DIM+1];
	unsigned latestPosChecked;
};

class SingleEndHeuristicIndelVerifier: public SingleEndIndelVerifier {
public:
	SingleEndHeuristicIndelVerifier(Indexer* indexer, unsigned distThresh) :
		SingleEndIndelVerifier(indexer, distThresh) {
	}

	void verifyCandidates(const std::string& read, std::vector<unsigned>& candidates,
			std::vector<ReadMapping>& results, bool unique);

	void verifyCandidates(unsigned& startPos, unsigned& k, const std::string& read,
			std::vector<unsigned>& candidates, std::vector<ReadMapping>& results);
};

#endif
