// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "globals.h"
#include "hobbesrunner.h"
#include "hobbesconf.h"
#include "hobbesstats.h"
#include "hobbesuserhelp.h"

#include <iostream>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <getopt.h>

using namespace std;

const char* optString = "-auk:q:p:n:v:f:i:r:c:";

static const struct option longOpts[] = {
		{ "all-mappings", no_argument, NULL, 'a' },
		{ "unique-mappings", no_argument, NULL, 'u' },
		{ "pe", no_argument, NULL, 0 },
		{ "k-mappings-per-read", required_argument, NULL, 'k' },
		{ "hamming", no_argument, NULL, 0 },
		{ "indel", no_argument, NULL, 0 },
		{ "min", required_argument, NULL, 0 },
		{ "max", required_argument, NULL, 0 },
		{ "allindel", no_argument, NULL, 0 },
		{ "reads-fastq-file", required_argument, NULL, 'q' },
		{ "seqfq1", required_argument, NULL, 0 },
		{ "seqfq2", required_argument, NULL, 0 },
		{ "pthreads", required_argument, NULL, 'p' },
		{ "num-reads", required_argument, NULL, 'n' },
		{ "distance", required_argument, NULL, 'v' },
		{ "reads-fasta-file", required_argument, NULL, 'f' },
		{ "indexfile", required_argument, NULL, 'i' },
		{ "reads-text-file", required_argument, NULL, 'r' },
		{ "read-sequence", required_argument, NULL, 'c' },
		{ "nofw", no_argument, NULL, 0 },
		{ "norc", no_argument, NULL, 0 },
		{ "sam-nohead", no_argument, NULL, 0 },
		{ "sam-nosq", no_argument, NULL, 0 },
		{ "sref", required_argument, NULL, 0 },
		{ "dref", required_argument, NULL, 0 },
		{ "mapout", required_argument, NULL, 0 },
		{ "stdin", no_argument, NULL, 0 },
		{ "gzip", no_argument, NULL, 0 },
		{ "bzip2", no_argument, NULL, 0 },
		{ "noprogress", no_argument, NULL, 0 },
		{ "version", no_argument, NULL, 0 },
		{ "help", no_argument, NULL, 0 },
		{ NULL, no_argument, NULL, 0 } };

bool check_numbers(const char* arg)
{
	for(int i = 0; arg[i] != '\0'; i++){
		if(arg[i] >= '0' && arg[i] <= '9') continue;
		else return false;
	}

	return true;
}

int atoi_safe(const char* arg, const char* msg)
{
	if(!check_numbers(arg)){
		cerr << "Invalid parameter value." << endl;
		cerr << "Usage: " << "--" << msg << " <int>" << endl;
		exit(1);
	}

	return atoi(arg);
}

int atoi_safe(const char* arg, int opt)
{
	if(!check_numbers(arg)){
		cerr << "Invalid parameter value." << endl;
		cerr << "Usage: " << "-" << char(opt) << " <int>" << endl;
		exit(1);
	}

	return atoi(arg);
}


int main(int argc, char **argv) {
	// Print help if no args given.
	if (argc == 1) {
		clog << user_help << endl;
		exit(0);
	}

	HobbesRunner runner;

	// Building command line input string.
	runner.conf.commandLineInput.clear();
	for (int i = 0; i < argc; i++) {
		runner.conf.commandLineInput.append(argv[i]);
		runner.conf.commandLineInput.append(" ");
	}

	vector<unsigned> complexOpts;
	int longIndex;
	int opt = getopt_long(argc, argv, optString, longOpts, &longIndex);
	while (opt != -1) {
		switch (opt) {
		case 'a':
			runner.conf.confOptions |= ALL_MAPPINGS;
			break;
		case 'u':
			runner.conf.confOptions |= ALL_MAPPINGS;
			runner.conf.uniqueMapping = true;
			break;
		case 'k':
			runner.conf.maxMappingsPerRead = atoi_safe(optarg, opt);
			break;
		case 'q':
			runner.conf.inputStreamType = FILE_INPUT;
			runner.conf.inputFormat = READS_FASTQ_FORMAT;
			runner.conf.readsFilename = optarg;
			break;
		case 'f':
			runner.conf.inputStreamType = FILE_INPUT;
			runner.conf.inputFormat = READS_FASTA_FORMAT;
			runner.conf.readsFilename = optarg;
			break;
		case 'c':
			runner.conf.inputStreamType = COMMAND_LINE_INPUT;
			runner.conf.inputFormat = READS_TXT_FORMAT;
			runner.conf.inputReadStream << optarg;
			break;
		case 'r':
			runner.conf.inputStreamType = FILE_INPUT;
			runner.conf.inputFormat = READS_TXT_FORMAT;
			runner.conf.readsFilename = optarg;
			break;
		case 'i':
			runner.conf.indexFileName = optarg;
			break;
		case 'p':
			runner.conf.numThreads = atoi_safe(optarg, opt);
			break;
		case 'n':
			runner.conf.numReads = atoi_safe(optarg, opt);
			if(runner.conf.numReads == 0){
				cerr << "Illegal number of reads." << endl;
				cerr << "Aborted" << endl;
				exit(1);
			}
			break;
		case 'v':
			runner.conf.distThresh = atoi_safe(optarg, opt);
			break;
		case 0:
			if (strcmp("nofw", longOpts[longIndex].name) == 0) {
				runner.conf.confOptions &= ~ONLY_REVERSE_ALIGN;
				runner.conf.confOptions |= ONLY_REVERSE_ALIGN;
			} else if (strcmp("norc", longOpts[longIndex].name) == 0) {
				runner.conf.confOptions &= ~ONLY_FORWARD_ALIGN;
				runner.conf.confOptions |= ONLY_FORWARD_ALIGN;
			} else if (strcmp("sref", longOpts[longIndex].name) == 0) {
				runner.conf.refSeqPathIsDir = false;
				runner.conf.refSeqPath = optarg;
			} else if (strcmp("dref", longOpts[longIndex].name) == 0) {
				runner.conf.refSeqPathIsDir = true;
				runner.conf.refSeqPath = optarg;
			} else if (strcmp("pe", longOpts[longIndex].name) == 0) {
				 // Paired-end uses both forward and reverse only.
				runner.conf.confOptions = ALL_MAPPINGS;
				runner.conf.readType = PAIRED;
			} else if (strcmp("seqfq1", longOpts[longIndex].name) == 0) {
				runner.conf.inputStreamType = FILE_INPUT;
				runner.conf.inputFormat = READS_PAIRED_FASTQ_FORMAT;
				runner.conf.readsFilename = optarg;
			} else if (strcmp("seqfq2", longOpts[longIndex].name) == 0) {
				runner.conf.secondReadsFilename = optarg;
			} else if (strcmp("hamming", longOpts[longIndex].name) == 0) {
				runner.conf.queryType = SELECT_SUBST;
			} else if (strcmp("indel", longOpts[longIndex].name) == 0) {
				runner.conf.queryType = SELECT_HEURISTIC_INDEL;
			} else if (strcmp("min", longOpts[longIndex].name) == 0) {
				runner.conf.minInsertSize = atoi_safe(optarg, "min");
			} else if (strcmp("max", longOpts[longIndex].name) == 0) {
				runner.conf.maxInsertSize = atoi_safe(optarg, "max");
			} else if (strcmp("allindel", longOpts[longIndex].name) == 0) {
				runner.conf.queryType = SELECT_INDEL;
			} else if (strcmp("sam-nohead", longOpts[longIndex].name) == 0) {
				runner.conf.samOutputOptions = SAM_NO_HEADERS;
			} else if (strcmp("sam-nosq", longOpts[longIndex].name) == 0) {
				runner.conf.samOutputOptions = SAM_NO_SEQ;
			} else if (strcmp("mapout", longOpts[longIndex].name) == 0) {
				runner.conf.outputStreamType = FILE_OUTPUT;
				runner.conf.unmappedOutputFileName = "UNMAP";
				runner.conf.mappedOutputFileName = optarg;
			} else if (strcmp("stdin", longOpts[longIndex].name) == 0) {
				runner.conf.inputStreamType = STANDARD_INPUT;
			} else if (strcmp("gzip", longOpts[longIndex].name) == 0) {
				runner.conf.inputCompression = GZIP;
			} else if (strcmp("bzip2", longOpts[longIndex].name) == 0) {
				runner.conf.inputCompression = BZIP2;
			} else if (strcmp("noprogress", longOpts[longIndex].name) == 0) {
				runner.conf.useProgressIndicator = false;
			} else if (strcmp("version", longOpts[longIndex].name) == 0) {
				clog << HOBBES_VERSION << " [" << HOBBES_PROGRAM_NAME << "]" << endl;
				exit(0);
			} else if (strcmp("help", longOpts[longIndex].name) == 0) {
				clog << user_help << endl;
				exit(0);
			}
			break;
		default: {
			abort();
		}
			break;

		}
		opt = getopt_long(argc, argv, optString, longOpts, &longIndex);
	}

	runner.conf.indexOptions = LOAD_INDEX;
	runner.conf.outputOptions |= WRITE_STATS;
	runner.conf.checkMappingConf();
	runner.run();

	return 0;
}
