// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "hobbesrunner.h"
#include "index/indexer.h"
#include "search/searcher.h"
#include "search/singleendsearcher.h"
#include "search/pairedendsearcher.h"
#include "input/linebyline.h"
#include "input/fasta.h"
#include "input/fastq.h"
#include "output/resultstringwriter.h"
#include "output/outputresultsink.h"
#include "tools/progressindicator.h"
#include "tools/statsutil.h"
#include "tools/looptimer.h"
#include "thread/threadstarter.h"
#include "thread/searchworker.h"
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/copy.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/iostreams/filter/bzip2.hpp>

#include <iostream>
#include <iomanip>
#include <fstream>
#include <istream>
#include <limits>

using namespace std;
using namespace boost::iostreams;

bool HobbesRunner::createInputStreams(istream** firstIn, istream** secondIn) {
	*firstIn = NULL;
	*secondIn = NULL;
	switch (conf.inputStreamType) {
	case COMMAND_LINE_INPUT: {
		*firstIn = &conf.inputReadStream;
		break;
	}
	case STANDARD_INPUT: {
		*firstIn = &cin;
		break;
	}
	case FILE_INPUT: {
		switch (conf.inputCompression) {
		case UNCOMPRESSED: {
			*firstIn = new ifstream(conf.readsFilename.c_str(), ios_base::in | ios_base::binary);
			if(!((ifstream*)*firstIn)->is_open()) return false;

			if (!conf.secondReadsFilename.empty()) {
				*secondIn = new ifstream(conf.secondReadsFilename.c_str(),
						ios_base::in | ios_base::binary);
				if(!((ifstream*)*secondIn)->is_open()) return false;
			}
			break;
		}
		case GZIP:
		case BZIP2: {
			zipFirstIn = new ifstream(conf.readsFilename.c_str(), ios_base::in | ios_base::binary);
			if(!zipFirstIn->is_open()) return false;

			filtering_istream* fin = new filtering_istream();
			if (conf.inputCompression == GZIP) {
				fin->push(gzip_decompressor());
			} else {
				fin->push(bzip2_decompressor());
			}
			fin->push(*zipFirstIn);
			*firstIn = fin;
			if (!conf.secondReadsFilename.empty()) {
				zipSecondIn = new ifstream(conf.secondReadsFilename.c_str(),
						ios_base::in | ios_base::binary);
				if(!zipSecondIn->is_open()) return false;

				filtering_istream* sin = new filtering_istream();
				if (conf.inputCompression == GZIP) {
					sin->push(gzip_decompressor());
				} else {
					sin->push(bzip2_decompressor());
				}
				sin->push(*zipSecondIn);
				*secondIn = sin;
			}
			break;
		}
		}
		break;
	}
	default: {
	}
	}

	return true;
}

InputReadQueue* HobbesRunner::createInputReadQueue(istream* firstIn, istream* secondIn) {
	switch (conf.inputFormat) {
	case READS_TXT_FORMAT:
		return new LineByLineInputReadQueue(*firstIn, conf.numReads, conf.inputMinOutstanding,
				conf.inputMaxOutstanding);
	case READS_FASTQ_FORMAT:
		return new FastqInputReadQueue(*firstIn, conf.numReads, conf.inputMinOutstanding,
				conf.inputMaxOutstanding);
	case READS_PAIRED_FASTQ_FORMAT:
		return new PairedEndFastqInputReadQueue(*firstIn, *secondIn, conf.numReads,
				conf.inputMinOutstanding, conf.inputMaxOutstanding);
	case READS_FASTA_FORMAT:
		return new FastaInputReadQueue(*firstIn, conf.numReads, conf.inputMinOutstanding,
				conf.inputMaxOutstanding);
	default:
		return NULL;
	}
}

ResultStringWriter* HobbesRunner::createResultStringWriter() {
	switch (conf.readType) {
	case SINGLE: 
		return new SAMFormatResultStringWriter(conf);
	
	case PAIRED: 
		return new PairedSAMFormatResultStringWriter(conf);
		break;
	}
	return NULL;
}

void HobbesRunner::createOutputStream(ostream** mappedOut) {
	switch (conf.outputStreamType) {
	case STANDARD_OUTPUT: {
		*mappedOut = &cout;
		break;
	}
	case FILE_OUTPUT: {
		*mappedOut = new ofstream(conf.mappedOutputFileName.c_str(), ios::trunc | ios::out);
		break;
	}
	}
}

OutputResultSink* HobbesRunner::createOutputResultSink(ostream* mappedOut) {
	switch (conf.outputResultSinkMode) {
	case ORSM_SYNC:
		return new SyncOutputResultSink(*mappedOut);
	case ORSM_ASYNC_FIFO:
		return new AsyncFifoOutputResultSink(*mappedOut, conf.outputMaxOutstanding);
	case ORSM_ASYNC_INORDER:
		return new AsyncInOrderOutputResultSink(*mappedOut, conf.outputMaxOutstanding);
	}
	return NULL;
}

Searcher* HobbesRunner::createSearcher(Indexer* indexer) {
	switch (conf.readType) {
	case SINGLE: {
		switch (conf.queryType) {
		case (SELECT_SUBST):
			return new SingleEndHammingSearcher(indexer, conf.distThresh, conf, stats);
		case (SELECT_INDEL):
			return new SingleEndIndelSearcher(indexer, conf.distThresh, conf, stats);
		case (SELECT_HEURISTIC_INDEL):
			return new SingleEndHeuristicIndelSearcher(indexer, conf.distThresh, conf, stats);
		default:
			clog << "ERROR: INVALID QUERY TYPE." << endl;
			return NULL;
		}
		break;
	}
	case PAIRED: {
		switch (conf.queryType) {
		case (SELECT_SUBST):
			return new PairedEndHammingSearcher(indexer, conf.distThresh, conf, stats);
		case (SELECT_INDEL):
			return new PairedEndIndelSearcher(indexer, conf.distThresh, conf, stats);
		case (SELECT_HEURISTIC_INDEL):
			return new PairedEndHeuristicIndelSearcher(indexer, conf.distThresh, conf, stats);
		default:
			clog << "ERROR: INVALID QUERY TYPE." << endl;
			return NULL;
		}
		break;
	}
	}
	return NULL;
}

void HobbesRunner::createIndex(){
	Indexer* indexer = new Indexer(conf, stats);

	// Check if reference sequence path is a file or a directory.
	if(conf.refSeqPathIsDir) indexer->readRefSeqFromDir();
	else indexer->readRefSeqFromFile(conf.refSeqPath);

	clog << "Reference sequence length: " << indexer->refSeq->size() << endl;
	stats.refSeqLen = indexer->refSeq->size();

	indexer->buildIndex();

	// indexer is not deleted here since the program will immediately exit
	return;
}

void HobbesRunner::run() {
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

	if(conf.indexOptions & CREATE_INDEX){
		createIndex();
		clog << "Index has been built and written to: " << conf.indexFileName << endl;
		exit(0);
	}

	// Create read input streams.
	istream* firstIn;
	istream* secondIn;
	if(!createInputStreams(&firstIn, &secondIn)){
		cerr << "Fatal error. Failed to open input file" << endl;
		exit(1);
	}

	// Create read input queue.
	InputReadQueue* inReadQueue = createInputReadQueue(firstIn, secondIn);

	// Retrieve the length of the first read
	int length = inReadQueue->estimateReadLength();
	if(length < 0){
		cerr << "Fatal error. Failed to read input stream" << endl;
		exit(1);
	}
	conf.readLength = unsigned(length);

	// Create indexer
	Indexer* indexer = new Indexer(conf, stats);

	// Check if reference sequence path is a file or a directory.
	if(conf.refSeqPathIsDir) indexer->readRefSeqFromDir();
	else indexer->readRefSeqFromFile(conf.refSeqPath);

	clog << "Reference sequence length: " << indexer->refSeq->size() << endl;
	stats.refSeqLen = indexer->refSeq->size();

	if((conf.indexOptions & LOAD_INDEX) == 0) {
		clog << "Fatal error: wrong index option" << endl;
		exit(1);
	}

	indexer->loadIndex();

	// numThreads for mappers, 1 thread for input queue, 1 thread for progress indicator.
	pthread_t inputReadQueueThreadHandle;
	pthread_t outputResultSinkHandle;
	pthread_t progressIndicatorThreadHandle;
	pthread_t mapperThreadHandles[conf.numThreads];

	unsigned progress_count = conf.numReads;
	if(progress_count == numeric_limits<unsigned>::max())
		progress_count = 0;
	ProgressIndicator progressIndicator(progress_count);
	ProgressIndicator* progIndPtr = NULL;
	if(conf.useProgressIndicator) {
		progIndPtr = &progressIndicator;
	}

	// Create result string writer.
	ResultStringWriter* resultWriter = createResultStringWriter();

	// Create output stream.
	ostream* mappedOut;
	createOutputStream(&mappedOut);

	// Create output result sink.
	OutputResultSink* outSink = createOutputResultSink(mappedOut);

	// Initialize output result sink.
	stringstream initialStream;
	resultWriter->createInitialString(initialStream, *indexer);
	outSink->writeInitialString(&initialStream);

	// Create one searcher instance per mapping thread.
	SearchWorker workers[conf.numThreads];
	for(unsigned i = 0; i < conf.numThreads; i++) {
		Searcher* searcher = createSearcher(indexer);
		workers[i].set(searcher, resultWriter, inReadQueue, outSink, progIndPtr);
	}

	StatsUtil sutil;
	sutil.startTimeMeasurement();

	// Start all threads.
	if(pthread_create(&inputReadQueueThreadHandle, &attr, startInputReadQueue, inReadQueue)) {
		cerr << "Fatal error. Failed to create input thread." << endl;
		exit(0);
	}
	if(conf.outputResultSinkMode == ORSM_ASYNC_FIFO || conf.outputResultSinkMode == ORSM_ASYNC_INORDER) {
		if(pthread_create(&outputResultSinkHandle, &attr, startOutputResultSink, outSink)) {
			cerr << "Fatal error. Failed to create progress indicator thread." << endl;
			exit(0);
		}
	}


	if(conf.useProgressIndicator) {
		if(pthread_create(&progressIndicatorThreadHandle, &attr, startMappingProgressIndicator, &progressIndicator)) {
			cerr << "Fatal error. Failed to create progress indicator thread." << endl;
			exit(0);
		}
	}


	for(unsigned i = 0; i < conf.numThreads; i++) {
		if(pthread_create(&mapperThreadHandles[i], &attr, startSearchWorker, &workers[i])) {
			cerr << "Fatal error. Failed to create mapper thread " << i << "." << endl;
			exit(0);
		}
	}

	// Join all threads.
	void* inrqStatus = 0;
	pthread_join(inputReadQueueThreadHandle, &inrqStatus);
	for(unsigned i = 0; i < conf.numThreads; i++) {
		void* status = 0;
		pthread_join(mapperThreadHandles[i], &status);
	}

	if(conf.outputResultSinkMode == ORSM_ASYNC_FIFO || conf.outputResultSinkMode == ORSM_ASYNC_INORDER) {
		AsyncOutputResultSink* asyncOutSink = dynamic_cast<AsyncOutputResultSink*>(outSink);
		asyncOutSink->terminate();
		void* status = 0;
		pthread_join(outputResultSinkHandle, &status);
	}

	unsigned long totalResults = 0;
	unsigned long totalSingleMappedResults = 0;
	// Note the "()" for initializing to 0.
	unsigned long* totalindel = new unsigned long[2 * conf.distThresh + 1]();
	unsigned long* totalmismatch = new unsigned long[2 * conf.distThresh + 1]();
	for (unsigned i = 0; i < conf.numThreads; i++) {
		if (workers[i].searcher) {
			totalResults += workers[i].searcher->getTotalMappings();
			totalSingleMappedResults += workers[i].searcher->getTotalSingleMappings();
			stats.numReadsMapped += workers[i].searcher->getNumMappedReads();
			stats.numReadsSingleMapped += workers[i].searcher->getNumSingleMappedReads();
			for (unsigned j = 0; j <= 2 * conf.distThresh; j++) {
				totalindel[j] += workers[i].searcher->indels[j];
				totalmismatch[j] += workers[i].searcher->mismatches[j];
			}
			delete workers[i].searcher;
		}
	}

	if(conf.useProgressIndicator) {
		progIndPtr->terminate(inReadQueue->getReadsDequeued());
		void* status = 0;
		pthread_join(progressIndicatorThreadHandle, &status);
	}

	sutil.stopTimeMeasurement();

	double ms = sutil.getTimeMeasurement(TFSEC);
	clog << "ELAPSED TIME: " << ms << "s" << endl;

	stats.numTotalMappings = totalResults;
	stats.numTotalSingleMappings = totalSingleMappedResults;

	stats.mergeTime /= conf.numReads * conf.numReads;
	stats.verificationTime /= conf.numReads * conf.numReads;
	stats.totalTime /= conf.numReads * conf.numReads;

	if (conf.outputOptions & WRITE_STATS) {
		unsigned readsDequeued = inReadQueue->getReadsDequeued();
		writeStats(stats.numReadsMapped, stats.numReadsSingleMapped, readsDequeued, totalResults,
				totalSingleMappedResults);
	}
	writeStatsSummary(clog, totalindel, totalmismatch);
	delete[] totalindel;
	delete[] totalmismatch;

	cleanupInputStreams(firstIn, secondIn);
	cleanupOutputStream(mappedOut);

	delete inReadQueue;
	delete resultWriter;
	delete outSink;
	delete indexer;
}

void HobbesRunner::cleanupInputStreams(istream* firstIn, istream* secondIn) {
	switch (conf.inputStreamType) {
	case FILE_INPUT: {
		switch (conf.inputCompression) {
		case UNCOMPRESSED: {
			if (firstIn) {
				ifstream* concreteStream = dynamic_cast<ifstream*>(firstIn);
				concreteStream->close();
				delete firstIn;
			}
			if (secondIn) {
				ifstream* concreteStream = dynamic_cast<ifstream*>(secondIn);
				concreteStream->close();
				delete secondIn;
			}
			break;
		}
		case GZIP:
		case BZIP2: {
			if (zipFirstIn) {
				zipFirstIn->close();
				delete zipFirstIn;
			}
			if (zipSecondIn) {
				zipSecondIn->close();
				delete zipSecondIn;
			}
			break;
		}
		}
		break;
	}
	default: {
		// Do nothing.
		break;
	}
	}
}

void HobbesRunner::cleanupOutputStream(ostream* mappedOut) {
	switch (conf.outputStreamType) {
	case FILE_OUTPUT:
		if (mappedOut) {
			ofstream* concreteStream = dynamic_cast<ofstream*>(mappedOut);
			concreteStream->close();
			delete mappedOut;
		}
		break;
	default:
		// Do nothing.
		break;
	}
}

void HobbesRunner::writeStats(unsigned numMappedReads,unsigned singleMappedReads, unsigned numReadsProcessed, unsigned numTotalMappings, unsigned numTotalSingleMappings) {
	clog << "#Reads processed : " << numReadsProcessed << endl;
	clog << "#Reads with atleast one mapping : " << numMappedReads << endl;
	clog << "#Reads with no mapping : " << (numReadsProcessed - numMappedReads) << endl;
	clog << "#Total mappings : " << numTotalMappings << endl;
	double refSeqGbs = (double)stats.refSeqLen / (1024.0 * 1024.0 * 1024.0);
	double indexGbs = (double)stats.actualIndexMem / (1024.0 * 1024.0 * 1024.0);
	double totalGbs = refSeqGbs + indexGbs;
	clog << "Approx. Peak Memory (ref sequence + index): " << setprecision(2) <<
			refSeqGbs << "GB + " << indexGbs << "GB = " << totalGbs << "GB" << endl;
}

void HobbesRunner::writeStatsSummary(ostream& out, unsigned long* totalindel,
		unsigned long* totalmismatch) {
	out << " " << endl;

	if (conf.readType == PAIRED){
		for (unsigned i = 0; i <= 2 * conf.distThresh; i++)
			if (totalmismatch[i] != 0)
				out << "#Reads within " << i << " mismatches: " << totalmismatch[i] << endl;
		for (unsigned i = 0; i <= 2 * conf.distThresh; i++) 
			if (totalindel[i] != 0) 
				out << "#Reads with " << i << " insertion/deletions: " << totalindel[i] << endl;
	} else {
		if (conf.queryType == SELECT_SUBST) {
			for (unsigned i = 0; i <= conf.distThresh; i++) 
				if (totalmismatch[i] != 0) 
					out << "#Reads within " << i << " mismatches: " << totalmismatch[i] << endl;
		} else if (conf.queryType == SELECT_HEURISTIC_INDEL) {
			for (unsigned i = 0; i <= conf.distThresh; i++) 
				if (totalmismatch[i] != 0) 
					out << "#Reads within " << i << " mismatches: " << totalmismatch[i] << endl;
			for (unsigned i = 0; i <= conf.distThresh; i++) 
				if (totalindel[i] != 0) 
					out << "#Reads with " << i << " insertion/deletions: " << totalindel[i] << endl;
		}
	}
}
