// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "output/outputresultsink.h"
#include <iostream>
#include <sstream>

using namespace std;

void OutputResultSink::writeResult(MappingResult& result) {
	stringstream* ss = (stringstream*)result.stream;
	switch (result.type) {
	case MRT_MAPPED:
		//mappedOut << result.stream->rdbuf();
		mappedOut << ss->str();
		break;
	case MRT_HALFMAPPED:
		//mappedOut << result.stream->rdbuf();
		mappedOut << ss->str();
		break;
	case MRT_UNMAPPED:
		//mappedOut << result.stream->rdbuf();
		mappedOut << ss->str();
		break;
	default:
		cerr << "WARNING: TRIED TO WRITE INVALID RESULT. RESULT IGNORED." << endl;
		break;
	}
	resultsWritten++;
}

void SyncOutputResultSink::add(MappingResult& result) {
	pthread_mutex_lock(&outputMutex);
	writeResult(result);
	result.clear();
	pthread_mutex_unlock(&outputMutex);
}

// Meant to be run as a thread.
void AsyncOutputResultSink::flushResults() {
	resultsWritten = 0;
	MappingResult result;
	while(!done) {
		pthread_mutex_lock(&queueMutex);
		while(sleep()) {
			// Wake up possibly sleeping producer threads.
			pthread_cond_wait(&flushCond, &queueMutex);
		}
		// Assuming done == true implies getQueteSize() == 0.
		if(getQueueSize() > 0) {
			pop(result);
		}
		// Signal possibly sleeping producer thread that the queue has space now.
		pthread_cond_signal(&stallCond);
		pthread_mutex_unlock(&queueMutex);
		// These locks prevent concurrent writing to the stream when terminate() is called
		// note that terminate could get the queueMutex and corrupt the output stream.
		pthread_mutex_lock(&terminateMutex);
		if(result.isValid()) {
			writeResult(result);
			result.clear();
		}
		pthread_mutex_unlock(&terminateMutex);
	}
}

void AsyncOutputResultSink::terminate() {
	pthread_mutex_lock(&queueMutex);
	pthread_mutex_lock(&terminateMutex);
	// Purge the queue.
	MappingResult result;
	while(getQueueSize() > 0) {
		pop(result);
		if(result.isValid()) {
			writeResult(result);
			result.clear();
		}
	}
	done = true;
	pthread_cond_signal(&flushCond);
	pthread_mutex_unlock(&terminateMutex);
	pthread_mutex_unlock(&queueMutex);
}

void AsyncOutputResultSink::add(MappingResult& result) {
	pthread_mutex_lock(&queueMutex);
	while(getQueueSize() >= maxOutstanding) {
		pthread_cond_wait(&stallCond, &queueMutex);
	}
	push(result);
	if(wakeup()) pthread_cond_signal(&flushCond);
	pthread_mutex_unlock(&queueMutex);
}


void AsyncFifoOutputResultSink::push(MappingResult& result) {
	resultQueue.push(result);
}

void AsyncFifoOutputResultSink::pop(MappingResult& result) {
	result = resultQueue.front();
	resultQueue.pop();
}

void AsyncInOrderOutputResultSink::push(MappingResult& result) {
	resultQueue.push_back(result);
	push_heap(resultQueue.begin(), resultQueue.end());
}

void AsyncInOrderOutputResultSink::pop(MappingResult& result) {
	result = resultQueue.front();
	pop_heap(resultQueue.begin(), resultQueue.end());
	resultQueue.pop_back();
	// Assuming rids are ordered integers from 1..n.
	lastRid++;
}

bool AsyncInOrderOutputResultSink::sleep() const {
	if(done) {
		return false;
	}
	if(resultQueue.empty()) {
		return true;
	}
	// Check whether the element at the top of the heap is the next one in rid order.
	if(resultQueue.front().readNum != lastRid + 1) {
		return true;
	}
	return false;
}

bool AsyncInOrderOutputResultSink::wakeup() const {
	// Check whether the element at the top of the heap is the next one in rid order.
	if(resultQueue.front().readNum == lastRid + 1) {
		return true;
	} else {
		return false;
	}
}
