// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "search/pairedendverifier.h"

#include <algorithm>
#include <limits>
#include <boost/lexical_cast.hpp>

#include <iostream>
using namespace std;

void PairedEndVerifier::setMinMaxDist(unsigned readLen) {
	if (minInsertSize < readLen) {
		minInsertSize = readLen;
	}
	minDist = minInsertSize - readLen;
	maxDist = maxInsertSize - readLen;
}

void PairedEndVerifier::verifyCandidates(PairedEndRead& read, vector<unsigned>& candSet1,
		vector<unsigned>& candSet2, vector<PairedReadMapping>& results, bool unique) {
	const string& read1 = *read.first.readStr;
	const string& read2 = *read.second.readStr;

	unsigned readLen = read1.size(); // assuming both reads have the same length.

	string defaultCigar = boost::lexical_cast < string > (read1.size());
	defaultCigar.append("M");

	vector<PairedReadMapping> tmp_res;

	vector<unsigned>::const_iterator candSet1Iter = candSet1.begin();
	vector<unsigned>::const_iterator candSet2Iter = candSet2.begin();

	for(;candSet1Iter != candSet1.end(); candSet1Iter++){
		while(candSet1Iter != candSet1.end() && *candSet1Iter + maxDist < *candSet2Iter) candSet1Iter++;
		if(candSet1Iter == candSet1.end()) break;

		while(candSet2Iter != candSet2.end() && *candSet1Iter + minDist > *candSet2Iter) candSet2Iter++;
		if(candSet2Iter == candSet2.end()) break;

		unsigned currentCandidate1 = *candSet1Iter;

		for(vector<unsigned>::const_iterator tmpIter = candSet2Iter;
		    tmpIter != candSet2.end() && *tmpIter <= *candSet1Iter + maxDist; tmpIter++){
			unsigned currentCandidate2 = *tmpIter;

			__builtin_prefetch((const void*) (refSeqPtr + (*(candSet1Iter + CACHE_PREFETCH_STEP))), 0, 0);
			__builtin_prefetch((const void*) (refSeqPtr + (*(tmpIter + CACHE_PREFETCH_STEP))), 0, 0);


			unsigned refMatchLen1 = -1;
			int offset1 = 0;
			int offset2 = 0;
			unsigned distance1 = singleEndVerifier->getDistance(refSeqPtr + currentCandidate1,
					readLen, read1, offset1, refMatchLen1);

			if (distance1 > distThresh) continue; 

			unsigned distance2 = singleEndVerifier->getDistance(refSeqPtr + currentCandidate2,
					readLen, read2, offset2, refMatchLen1);

			if (distance2 > distThresh) continue;

			vector<unsigned>::const_iterator it = upper_bound(indexer->chromosomeLengths.begin(),
					indexer->chromosomeLengths.end(), currentCandidate1);
			if (currentCandidate2 + readLen <= *it) {
				if(updateStats) tempMisMatch[distance1 + distance2]++;

				if(unique && !tmp_res.empty()) return;
				tmp_res.push_back(
						PairedReadMapping(currentCandidate1 + offset1, distance1,
								  currentCandidate2 + offset2, distance2, defaultCigar, defaultCigar));
			}
		}
	}

	std::swap(tmp_res, results);
}

void PairedEndIndelVerifier::verifyCandidates(PairedEndRead& read, vector<unsigned>& candSet1,
		vector<unsigned>& candSet2, vector<PairedReadMapping>& results, bool unique) {
	SingleEndIndelVerifier* indelVerifier = static_cast<SingleEndIndelVerifier*>(singleEndVerifier);

	const string& read1 = *read.first.readStr;
	const string& read2 = *read.second.readStr;
	// Assuming both reads have the same length.
	unsigned readLen = read1.size();
	GramCountArray readGramCounts1;
	GramCountArray readGramCounts2;
	indelVerifier->fillGramCountArray(read1.data(), read1.length(), readGramCounts1);
	indelVerifier->fillGramCountArray(read2.data(), read2.length(), readGramCounts2);
	unsigned distance1 = numeric_limits<unsigned>::max();
	unsigned distance2 = numeric_limits<unsigned>::max();

	stringstream cigar1;
	stringstream cigar2;

	string defaultCigar = boost::lexical_cast < string > (read1.size());
	defaultCigar.append("M");

	vector<PairedReadMapping> tmp_res;

	unsigned minBound = minDist - distThresh;
	unsigned maxBound = maxDist + distThresh;

	vector<unsigned>::const_iterator candSet1Iter = candSet1.begin();
	vector<unsigned>::const_iterator candSet2Iter = candSet2.begin();

	for(;candSet1Iter != candSet1.end(); candSet1Iter++){
		while(candSet1Iter != candSet1.end() && *candSet1Iter + maxBound < *candSet2Iter) candSet1Iter++;
		if(candSet1Iter == candSet1.end()) break;

		while(candSet2Iter != candSet2.end() && *candSet1Iter + minBound > *candSet2Iter) candSet2Iter++;
		if(candSet2Iter == candSet2.end()) break;

		unsigned currentCandidate1 = *candSet1Iter;
		for(vector<unsigned>::const_iterator tmpIter = candSet2Iter;
		    tmpIter != candSet2.end() && *tmpIter <= *candSet1Iter + maxBound; tmpIter++){
			unsigned currentCandidate2 = *tmpIter;

			__builtin_prefetch((const void*) (refSeqPtr + (*(candSet1Iter + CACHE_PREFETCH_STEP))), 0, 0);
			__builtin_prefetch((const void*) (refSeqPtr + (*(tmpIter + CACHE_PREFETCH_STEP))), 0, 0);

			unsigned refMatchLen1 = -1;
			unsigned refMatchLen2 = -1;

			unsigned maxLength = readLen + distThresh;

			unsigned lowerBound = 0;

			indelVerifier->getDistanceLowerBound(readGramCounts1, currentCandidate1, readLen);
			if(lowerBound > (distThresh + 1)) continue;

			indelVerifier->getDistanceLowerBound(readGramCounts2, currentCandidate2, readLen);
			if(lowerBound > distThresh + 1) continue;

			cigar1.str("");

			unsigned indel1=-1;
			int offset1 = 0;
			distance1= indelVerifier->getBitVectorEditDistance(currentCandidate1, refMatchLen1,
                                                                 maxLength, read1,distThresh, cigar1,indel1, offset1);
			if(distance1 > distThresh) continue;

			cigar2.str("");

			unsigned indel2 = -1;
			int offset2 = 0;
			distance2= indelVerifier->getBitVectorEditDistance(currentCandidate2, refMatchLen2,
                                                                maxLength, read2,distThresh, cigar2, indel2, offset2);
			if(distance2 > distThresh) continue;

			// check boundary once again for the offsets
			if((currentCandidate2 + offset2 < (currentCandidate1 + offset1 + minDist)) || 
			   (currentCandidate2 + offset2 > (currentCandidate1 + offset1 + maxDist))) 
				continue;

			vector<unsigned>::const_iterator it =
				upper_bound(indexer->chromosomeLengths.begin(),
							indexer->chromosomeLengths.end(), currentCandidate1 + offset1);
			vector<unsigned>::const_iterator it2 =
				upper_bound(indexer->chromosomeLengths.begin(),
							indexer->chromosomeLengths.end(), currentCandidate2 + offset2);
			if ((currentCandidate1 + offset1 + readLen <= *it)
					&& (currentCandidate2 + offset2 + readLen <= *it2)) {
				string chrName = indexer->chromosomeNames.at(
						(it - indexer->chromosomeLengths.begin()));
				string chrName1 = indexer->chromosomeNames.at(
						(it2 - indexer->chromosomeLengths.begin()));
				if (chrName == chrName1) {
					if(updateStats){
						if (indel1 > 0 || indel2 > 0) tempIndel[indel1 + indel2]++;
						else tempMisMatch[distance1 + distance2]++;
					}

					if(unique && !tmp_res.empty()) return;
					tmp_res.push_back(PairedReadMapping(currentCandidate1 + offset1,
									distance1, currentCandidate2 + offset2,
									distance2, cigar1.str(), cigar2.str()));
				}
			}
		}
	}

	std::swap(tmp_res, results);
}
