// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "include/globals.h"
#include "include/search/searcher.h"

#include <iostream>
#include <fstream>
#include <istream>

using namespace std;

void Searcher::findMappings(ReadInfo& readInfo){
	prepare(readInfo.getLength());

	if (conf.confOptions & ALL_MAPPINGS) {
		findAllMappings(readInfo);
	} else {
		findKMappings(readInfo);
	}

	// First = results size, second = candidates size.
	pair<unsigned, unsigned> counts;
	getStats(counts);
	totalMappings += counts.first;
	totalSingleMappings += counts.second;

	getBestStats();

	switch (getMappingResultType()){
	case MRT_MAPPED:
		numMappedReads++;
		break;
	case MRT_HALFMAPPED:
		numSingleMappedReads++;
		break;
	default:
		break;
	}
}

bool Searcher::fillRead(Read& mappingReadStruct, unsigned& tOccurThresh) {
	mappingReadStruct.getInvertedLists(indexer, gramGen);
	unsigned numGrams = mappingReadStruct.grams.size();
	tOccurThresh = (numGrams > distThresh * gramGen.gramLen) ? numGrams - distThresh * gramGen.gramLen : 0;

	unsigned invListSize = 0;
	if(indexer->useBitVector) invListSize = mappingReadStruct.bitvectorInvertedLists.size();
	else invListSize = mappingReadStruct.invertedLists.size();

	if ((invListSize < tOccurThresh) || (tOccurThresh == 0)) {
		return false;
	}

	// Hobbes2: additionalPrefix scheme is available for ALL_MAPPINGS with HAMMING and HEURISTIC_INDEL only
	mappingReadStruct.additionalPrefix = 0;
	if(invListSize > (distThresh + 1) * gramGen.gramLen &&
	   (conf.confOptions & ALL_MAPPINGS) && conf.queryType != SELECT_INDEL)
		mappingReadStruct.additionalPrefix = 1;

	if(indexer->useBitVector)
		mappingReadStruct.getPrefixDynamicProgrammingWithBitVector(distThresh,
									   mappingReadStruct.readStr->size(),
									   gramGen.gramLen);
	else
		mappingReadStruct.getPrefixDynamicProgrammingWithoutBitVector(distThresh,
									      mappingReadStruct.readStr->size(),
									      gramGen.gramLen);
	return true;
}

void Searcher::getReverseComplementRead(string& read) {
	unsigned strLen = read.size();
	for (unsigned i = 0; i < (strLen / 2); i++) {
		complementChar(read[i]);
		complementChar(read[(strLen - 1) - i]);
		swap(read[i], read[(strLen - 1) - i]);
	}
	// For odd read lengths.
	if ((strLen % 2) != 0) {
		complementChar(read[(strLen / 2)]);
	}
}

void Searcher::getReverseReadQuality(string& readQuality) {
	unsigned strLen = readQuality.size();
	for (unsigned i = 0; i < (strLen / 2); i++) {
		swap(readQuality[i], readQuality[(strLen - 1) - i]);
	}
}

void Searcher::complementChar(char& c) {
	switch (c) {
	case BASE_A:
		c = BASE_T;
		break;
	case BASE_T:
		c = BASE_A;
		break;
	case BASE_G:
		c = BASE_C;
		break;
	case BASE_C:
		c = BASE_G;
		break;
	default:
		break;
	}
}
