// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "search/singleendmerger.h"

#include <iostream>
#include <cstdlib>
#include <limits>

using namespace std;

// ===================================================================================
// In this file, we implemented the following three member functions of each class.
// All of them do not use bit vectors for filtering candidates.
//
// The first function generate candidates using k+2 scheme with bit vector inverted lists.
// The second function generate candidates using k+2 scheme with normal inverted lists.
// The third function generates candidates using k+1 scheme with normal inverted lists.
// ===================================================================================

// ============================================================================
// SingleEndIndelMerger
// ============================================================================
void SingleEndIndelMerger::findCandidatesAll(const vector<BitVectorInvertedListWithOffset>& invLists,
					     const vector<unsigned>& prefixGrams,
					     vector<unsigned>& candidates) 
{ 
	clog << endl << "NOT IMPLEMENTED YET!!" << endl;
	exit(1);
	
	// not supported yet!!
	// will be implemented soon
}

void SingleEndIndelMerger::findCandidatesAll(const vector<InvertedListWithOffset>& invLists,
					     const vector<unsigned>& prefixGrams,
					     vector<unsigned>& candidates) 
{ 
	// not supported yet!!
	// will be implemented soon
}

void SingleEndIndelMerger::findCandidatesAllNF( const vector<InvertedListWithOffset>& invLists,
						const vector<unsigned>& prefixGrams,
						vector<unsigned>& candidates)
{ 
	// Panic case. Return nothing.
	if ((invLists.size() < distThresh) || (distThresh == 0)) {
		return;
	}
	vector<unsigned>::const_iterator invListIter = invLists[prefixGrams[0]].second->begin();
	vector<unsigned>::const_iterator invListEnd = invLists[prefixGrams[0]].second->end();
	unsigned gramPosition = invLists[prefixGrams[0]].first;
	tmp.clear();
	while (invListIter != invListEnd) {
		if (*invListIter >= gramPosition) {
			break;
		}
		invListIter++;
	}
	while (invListIter != invListEnd) {
		candidates.push_back((*invListIter - gramPosition));
		invListIter++;
	}
	for (unsigned j = 1; j < prefixGrams.size(); j++) {
		unsigned i = prefixGrams[j];
		unsigned gramPosition = invLists[i].first;
		tmp.clear();
		tmp.reserve(candidates.size() + invLists[i].second->size());
		vector<unsigned>::iterator candIter = candidates.begin();
		invListIter = invLists[i].second->begin();
		invListEnd = invLists[i].second->end();
		while (invListIter != invListEnd) {
			if (*invListIter >= gramPosition) {
				break;
			}
			invListIter++;
		}
		while (candIter != candidates.end()) {
			if ((invListIter != invListEnd && *candIter > (*invListIter - gramPosition))) {
				tmp.push_back((*invListIter - gramPosition));
				invListIter++;
			} else if (invListIter == invListEnd
					|| (*candIter < (*invListIter - gramPosition))) {
				tmp.push_back(*candIter++);
			} else {
				tmp.push_back(*candIter++);
				invListIter++;
			}
		}
		while (invListIter < invListEnd) {
			tmp.push_back((*invListIter - gramPosition));
			invListIter++;
		}
		std::swap(candidates, tmp);
	}
}

// ============================================================================
// SingleEndHeuristicIndelMerger
// ============================================================================
void SingleEndHeuristicIndelMerger::findCandidatesAll(const vector<BitVectorInvertedListWithOffset>& invLists,
						      const vector<unsigned>& prefixGrams,
						      vector<unsigned>& candidates) 
{
	vector<BitVecPair>::const_iterator invListIter = invLists[prefixGrams[0]].second->begin();
	vector<BitVecPair>::const_iterator invListEnd = invLists[prefixGrams[0]].second->end();
	unsigned gramPosition = invLists[prefixGrams[0]].first;

	typedef pair<unsigned, unsigned> c_type;
	vector<c_type> cands;

	while (invListIter != invListEnd) {
		if (invListIter->pos >= gramPosition) {
			break;
		}
		invListIter++;
	}
	while (invListIter != invListEnd) {
		cands.push_back(c_type(invListIter->pos - gramPosition, 1));
		invListIter++;
	}

	//unsigned tmpDelta = delta;
	//if(prefixGrams.size() <= distThresh + 1) delta = 0;

	vector<c_type> temp;
	for (unsigned j = 1; j < prefixGrams.size(); j++) {
		unsigned i = prefixGrams[j];
		unsigned gramPosition = invLists[i].first;

		temp.clear();
		temp.reserve(cands.size() + invLists[i].second->size());
		vector<c_type>::const_iterator candIter = cands.begin();
		invListIter = invLists[i].second->begin();
		invListEnd = invLists[i].second->end();
		while (invListIter != invListEnd) {
			if (invListIter->pos >= gramPosition) {
				break;
			}
			invListIter++;
		}

		int additionalPrefix =  j - distThresh;
		int carry = 0;

		while (candIter != cands.end()) {
			if (invListIter != invListEnd && rgt(candIter->first,  invListIter->pos - gramPosition)) {
			    if(additionalPrefix <= 0 || carry != 0){
				temp.push_back(c_type(invListIter->pos - gramPosition, 1 + carry));
				carry = 0;
			    }
			    invListIter++;
			} else if (invListIter == invListEnd || rlt(candIter->first, invListIter->pos - gramPosition)) {
				if(additionalPrefix <= 0 || candIter->second + carry > unsigned(additionalPrefix)){
					temp.push_back(c_type(candIter->first, candIter->second + carry));
					carry = 0;
				}
				candIter++;
			} else { 
				if(candIter->first == (invListIter->pos - gramPosition)){
					temp.push_back(c_type(candIter->first, candIter->second + 1));
					candIter++;
					invListIter++;
					carry = 0;
				}
				else if(candIter->first > (invListIter->pos - gramPosition)){
					temp.push_back(c_type(invListIter->pos - gramPosition, 2));
					invListIter++;
					carry = 1;
				}
				else{
					temp.push_back(c_type(candIter->first, candIter->second + 1));
					candIter++;
					carry = 1;
				}
			}
		}
		while (invListIter < invListEnd && additionalPrefix <= 0) {
			temp.push_back(c_type(invListIter->pos - gramPosition, 1));

			invListIter++;
		}
		std::swap(cands, temp);
	}

	//delta = tmpDelta;

	for(unsigned i = 0; i < cands.size(); i++){
		if(i == 0 || candidates.back() != cands[i].first) // remove dups if any
			candidates.push_back(cands[i].first);
	}
}

// reduced memory version
void SingleEndHeuristicIndelMerger::findCandidatesAll(const vector<InvertedListWithOffset>& invLists,
						      const vector<unsigned>& prefixGrams,
						      vector<unsigned>& candidates) 
{
	vector<unsigned>::const_iterator invListIter = invLists[prefixGrams[0]].second->begin();
	vector<unsigned>::const_iterator invListEnd = invLists[prefixGrams[0]].second->end();
	unsigned gramPosition = invLists[prefixGrams[0]].first;

	typedef pair<unsigned, unsigned> c_type;
	vector<c_type> cands;

	while (invListIter != invListEnd) {
		if (*invListIter >= gramPosition) {
			break;
		}
		invListIter++;
	}
	while (invListIter != invListEnd) {
		cands.push_back(c_type(*invListIter - gramPosition, 1));
		invListIter++;
	}

	//unsigned tmpDelta = delta;
	//if(prefixGrams.size() <= distThresh + 1) delta = 0;

	vector<c_type> temp;
	for (unsigned j = 1; j < prefixGrams.size(); j++) {
		unsigned i = prefixGrams[j];
		unsigned gramPosition = invLists[i].first;

		temp.clear();
		temp.reserve(cands.size() + invLists[i].second->size());
		vector<c_type>::const_iterator candIter = cands.begin();
		invListIter = invLists[i].second->begin();
		invListEnd = invLists[i].second->end();
		while (invListIter != invListEnd) {
			if (*invListIter >= gramPosition) {
				break;
			}
			invListIter++;
		}

		int additionalPrefix =  j - distThresh;
		int carry = 0;

		while (candIter != cands.end()) {
			if (invListIter != invListEnd && rgt(candIter->first,  *invListIter - gramPosition)) {
			    if(additionalPrefix <= 0 || carry != 0){
				temp.push_back(c_type(*invListIter - gramPosition, 1 + carry));
				carry = 0;
			    }
			    invListIter++;
			} else if (invListIter == invListEnd || rlt(candIter->first, *invListIter - gramPosition)) {
				if(additionalPrefix <= 0 || candIter->second + carry > unsigned(additionalPrefix)){
					temp.push_back(c_type(candIter->first, candIter->second + carry));
					carry = 0;
				}
				candIter++;
			} else { 
				if(candIter->first == (*invListIter - gramPosition)){
					temp.push_back(c_type(candIter->first, candIter->second + 1));
					candIter++;
					invListIter++;
					carry = 0;
				}
				else if(candIter->first > (*invListIter - gramPosition)){
					temp.push_back(c_type(*invListIter - gramPosition, 2));
					invListIter++;
					carry = 1;
				}
				else{
					temp.push_back(c_type(candIter->first, candIter->second + 1));
					candIter++;
					carry = 1;
				}
			}
		}
		while (invListIter < invListEnd && additionalPrefix <= 0) {
			temp.push_back(c_type(*invListIter - gramPosition, 1));

			invListIter++;
		}
		std::swap(cands, temp);
	}

	//delta = tmpDelta;

	for(unsigned i = 0; i < cands.size(); i++){
		if(i == 0 || candidates.back() != cands[i].first) // remove dups if any
			candidates.push_back(cands[i].first);
	}
}

void SingleEndHeuristicIndelMerger::findCandidatesAllNF(const vector<InvertedListWithOffset>& invLists,
							const vector<unsigned>& prefixGrams,
							vector<unsigned>& candidates)
{
	vector<unsigned>::const_iterator invListIter = invLists[prefixGrams[0]].second->begin();
	vector<unsigned>::const_iterator invListEnd = invLists[prefixGrams[0]].second->end();
	unsigned gramPosition = invLists[prefixGrams[0]].first;

	while (invListIter != invListEnd) {
		if (*invListIter >= gramPosition) {
			break;
		}
		invListIter++;
	}
	while (invListIter != invListEnd) {
		candidates.push_back(*invListIter - gramPosition);
		invListIter++;
	}

	vector<unsigned> temp;
	for (unsigned j = 1; j < prefixGrams.size(); j++) {
		unsigned i = prefixGrams[j];
		unsigned gramPosition = invLists[i].first;

		temp.clear();
		temp.reserve(candidates.size() + invLists[i].second->size());
		vector<unsigned>::const_iterator candIter = candidates.begin();
		invListIter = invLists[i].second->begin();
		invListEnd = invLists[i].second->end();
		while (invListIter != invListEnd) {
			if (*invListIter >= gramPosition) {
				break;
			}
			invListIter++;
		}

		while (candIter != candidates.end()) {
			if (invListIter != invListEnd && *candIter >  *invListIter - gramPosition) {
				temp.push_back(*invListIter - gramPosition);
				invListIter++;
			} else if (invListIter == invListEnd || *candIter < *invListIter - gramPosition) {
				temp.push_back(*candIter);
				candIter++;
			} else { 
				temp.push_back(*candIter);
				candIter++;
				invListIter++;
			}
		}
		while (invListIter < invListEnd){
			temp.push_back(*invListIter - gramPosition);
			invListIter++;
		}
		std::swap(candidates, temp);
	}

/*
	for(unsigned i = 0; i < cands.size(); i++){
		if(i == 0 || candidates.back() != cands[i].first) // remove dups if any
			candidates.push_back(cands[i].first);
	}
*/
}


// ============================================================================
// SingleEndHammingMerger
// ============================================================================
void SingleEndHammingMerger::findCandidatesAll(const vector<BitVectorInvertedListWithOffset>& invLists,
					       const vector<unsigned>& prefixGrams,
					       vector<unsigned>& candidates) 
{
	vector<BitVecPair>::const_iterator invListIter = invLists[prefixGrams[0]].second->begin();
	vector<BitVecPair>::const_iterator invListEnd = invLists[prefixGrams[0]].second->end();
	unsigned gramPosition = invLists[prefixGrams[0]].first;

        typedef pair<unsigned, unsigned> c_type;
        vector<c_type> cands;
        
	while (invListIter != invListEnd) {
		if (invListIter->pos >= gramPosition) {
			break;
		}
		invListIter++;
	}
	while (invListIter != invListEnd) {
		cands.push_back(c_type(invListIter->pos - gramPosition, 1));
		invListIter++;
	}

        vector<c_type> temp;
	for (unsigned j = 1; j < prefixGrams.size(); j++) {
		unsigned i = prefixGrams[j];
		unsigned gramPosition = invLists[i].first;

		temp.clear();
		temp.reserve(cands.size() + invLists[i].second->size());
		vector<c_type>::const_iterator candIter = cands.begin();
		invListIter = invLists[i].second->begin();
		invListEnd = invLists[i].second->end();
		while (invListIter != invListEnd) {
			if (invListIter->pos >= gramPosition) {
				break;
			}
			invListIter++;
		}

                bool additionalPrefix = (j >= distThresh + 1);
                
		while (candIter != cands.end()) {
			if ((invListIter != invListEnd && candIter->first > (invListIter->pos - gramPosition))) {
				if (!additionalPrefix)
					temp.push_back(c_type(invListIter->pos - gramPosition, 1));
				invListIter++;
			} else if (invListIter == invListEnd
					|| (candIter->first < (invListIter->pos - gramPosition))) {
				if(!additionalPrefix || candIter->second > 1)
					temp.push_back(*candIter);
				candIter++;
			} else {
				temp.push_back(c_type(candIter->first, candIter->second+1));
				candIter++;
				invListIter++;
			}
		}

		while (invListIter < invListEnd && !additionalPrefix) {
			temp.push_back(c_type(invListIter->pos - gramPosition, 1));
			invListIter++;
		}

		std::swap(cands, temp);
	}

	for(unsigned i = 0; i < cands.size(); i++){
		if(i == 0 || candidates.back() != cands[i].first) // remove dups if any
			candidates.push_back(cands[i].first);
	}
}

// reduced memory version
void SingleEndHammingMerger::findCandidatesAll(const vector<InvertedListWithOffset>& invLists,
					       const vector<unsigned>& prefixGrams,
					       vector<unsigned>& candidates) 
{
	vector<unsigned>::const_iterator invListIter = invLists[prefixGrams[0]].second->begin();
	vector<unsigned>::const_iterator invListEnd = invLists[prefixGrams[0]].second->end();
	unsigned gramPosition = invLists[prefixGrams[0]].first;

        typedef pair<unsigned, unsigned> c_type;
        vector<c_type> cands;
        
	while (invListIter != invListEnd) {
		if (*invListIter >= gramPosition) {
			break;
		}
		invListIter++;
	}
	while (invListIter != invListEnd) {
		cands.push_back(c_type(*invListIter - gramPosition, 1));
		invListIter++;
	}

        vector<c_type> temp;
	for (unsigned j = 1; j < prefixGrams.size(); j++) {
		unsigned i = prefixGrams[j];
		unsigned gramPosition = invLists[i].first;

		temp.clear();
		temp.reserve(cands.size() + invLists[i].second->size());
		vector<c_type>::const_iterator candIter = cands.begin();
		invListIter = invLists[i].second->begin();
		invListEnd = invLists[i].second->end();
		while (invListIter != invListEnd) {
			if (*invListIter >= gramPosition) {
				break;
			}
			invListIter++;
		}

                bool additionalPrefix = (j >= distThresh + 1);
                
		while (candIter != cands.end()) {
			if ((invListIter != invListEnd && candIter->first > (*invListIter - gramPosition))) {
				if (!additionalPrefix)
					temp.push_back(c_type(*invListIter - gramPosition, 1));
				invListIter++;
			} else if (invListIter == invListEnd
					|| (candIter->first < (*invListIter - gramPosition))) {
				if(!additionalPrefix || candIter->second > 1)
					temp.push_back(*candIter);
				candIter++;
			} else {
				temp.push_back(c_type(candIter->first, candIter->second+1));
				candIter++;
				invListIter++;
			}
		}

		while (invListIter < invListEnd && !additionalPrefix) {
			temp.push_back(c_type(*invListIter - gramPosition, 1));
			invListIter++;
		}

		std::swap(cands, temp);
	}

	for(unsigned i = 0; i < cands.size(); i++){
		if(i == 0 || candidates.back() != cands[i].first) // remove dups if any
			candidates.push_back(cands[i].first);
	}
}

// Implementation of the following function is exactly the same as that of 
// SingleEndHeuristicIndelMerger::findCandidatesAllNF.
void SingleEndHammingMerger::findCandidatesAllNF(const vector<InvertedListWithOffset>& invLists,
						 const vector<unsigned>& prefixGrams,
						 vector<unsigned>& candidates)
{
	vector<unsigned>::const_iterator invListIter = invLists[prefixGrams[0]].second->begin();
	vector<unsigned>::const_iterator invListEnd = invLists[prefixGrams[0]].second->end();
	unsigned gramPosition = invLists[prefixGrams[0]].first;

	while (invListIter != invListEnd) {
		if (*invListIter >= gramPosition) {
			break;
		}
		invListIter++;
	}
	while (invListIter != invListEnd) {
		candidates.push_back(*invListIter - gramPosition);
		invListIter++;
	}

	vector<unsigned> temp;
	for (unsigned j = 1; j < prefixGrams.size(); j++) {
		unsigned i = prefixGrams[j];
		unsigned gramPosition = invLists[i].first;

		temp.clear();
		temp.reserve(candidates.size() + invLists[i].second->size());
		vector<unsigned>::const_iterator candIter = candidates.begin();
		invListIter = invLists[i].second->begin();
		invListEnd = invLists[i].second->end();
		while (invListIter != invListEnd) {
			if (*invListIter >= gramPosition) {
				break;
			}
			invListIter++;
		}

		while (candIter != candidates.end()) {
			if (invListIter != invListEnd && *candIter >  *invListIter - gramPosition) {
				temp.push_back(*invListIter - gramPosition);
				invListIter++;
			} else if (invListIter == invListEnd || *candIter < *invListIter - gramPosition) {
				temp.push_back(*candIter);
				candIter++;
			} else { 
				temp.push_back(*candIter);
				candIter++;
				invListIter++;
			}
		}
		while (invListIter < invListEnd){
			temp.push_back(*invListIter - gramPosition);
			invListIter++;
		}
		std::swap(candidates, temp);
	}

/*
	for(unsigned i = 0; i < cands.size(); i++){
		if(i == 0 || candidates.back() != cands[i].first) // remove dups if any
			candidates.push_back(cands[i].first);
	}
*/
}
