// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "search/singleendsearcher.h"

#include <cstring>

using namespace std;

void SingleEndSearcher::prepare(unsigned readLen) {
	forwardCandidates.clear();
	forwardResults.clear();
	reverseCandidates.clear();
	reverseResults.clear();

	verifier->updateStats = true;
	memset(verifier->tempIndel, 0, (2 * distThresh + 1) * sizeof(long));
	memset(verifier->tempMisMatch, 0, (2 * distThresh + 1) * sizeof(long));
}

void SingleEndSearcher::getStats(pair<unsigned, unsigned>& counts) {
	counts.first = forwardResults.size() + reverseResults.size();
	counts.second = forwardCandidates.size() + reverseCandidates.size();
}

void SingleEndSearcher::getBestStats() {
	if(forwardResults.empty() && reverseResults.empty()) return;

	if (verifier->tempMisMatch[0] > 0) {
		mismatches[0]++;
	} else {
		for (unsigned j = 1; j <= distThresh; j++) {
			if (verifier->tempMisMatch[j] != 0 || verifier->tempIndel[j] != 0) {
				if (verifier->tempMisMatch[j] >= verifier->tempIndel[j]) {
					mismatches[j]++;
				} else{
					indels[j]++;
				}
				break;
			}
		}
	}
}

void SingleEndSearcher::findAllMappings(ReadInfo& readInfo) {
	unsigned tOccurThresh = 0;
	if ((conf.confOptions & ONLY_FORWARD_ALIGN) || (conf.confOptions & ONLY_REVERSE_ALIGN)) {
		if (conf.confOptions & ONLY_FORWARD_ALIGN) {
			forwardRead.setReadString(&readInfo.readString);
		} else if (conf.confOptions & ONLY_REVERSE_ALIGN) {
			genReverses(readInfo);
			forwardRead.setReadString(&readInfo.reverseReadString);
		}
		if (fillRead(forwardRead, tOccurThresh)) {
			mergeAndVerifyAll(forwardRead, forwardCandidates, forwardResults);
		}
	} else {
		forwardRead.setReadString(&readInfo.readString);
		if (fillRead(forwardRead, tOccurThresh)) {
			mergeAndVerifyAll(forwardRead, forwardCandidates, forwardResults);
		}
		genReverses(readInfo);
		forwardRead.setReadString(&readInfo.reverseReadString);
		if (fillRead(forwardRead, tOccurThresh)) {
			if(conf.uniqueMapping && !forwardResults.empty())
				verifier->updateStats = false;
			mergeAndVerifyAll(forwardRead, reverseCandidates, reverseResults);
		}

		if(conf.uniqueMapping && !forwardResults.empty() && !reverseResults.empty()){
			forwardResults.clear();
			reverseResults.clear();
		}
	}
}

void SingleEndSearcher::findKMappings(ReadInfo& inputReadStruct) {
	unsigned tOccurThresh1 = 0;
	unsigned k = conf.maxMappingsPerRead;
	bool done = false;
	if ((conf.confOptions & ONLY_FORWARD_ALIGN) || (conf.confOptions & ONLY_REVERSE_ALIGN)) {
		if (conf.confOptions & ONLY_FORWARD_ALIGN) {
			forwardRead.setReadString(&inputReadStruct.readString);
		} else if (conf.confOptions & ONLY_REVERSE_ALIGN) {
			genReverses(inputReadStruct);
			forwardRead.setReadString(&inputReadStruct.reverseReadString);
		}
		// Find exact matches first.
		if (fillRead(forwardRead, tOccurThresh1)) {
			done = mergeAndVerifyK(true, k, forwardRead, forwardCandidates, forwardResults);
		} else {
			return;
		}
		if (!done) {
			mergeAndVerifyK(false, k, forwardRead, forwardCandidates, forwardResults);
		}
	} else {
		unsigned tOccurThresh2 = 0;
		forwardRead.setReadString(&inputReadStruct.readString);
		// Find exact matches first.
		if (fillRead(forwardRead, tOccurThresh1)) {
			done = mergeAndVerifyK(true, k, forwardRead, forwardCandidates, forwardResults);
		}
		if (!done) {
			genReverses(inputReadStruct);
			reverseRead.setReadString(&inputReadStruct.reverseReadString);
			// Find exact matches first - reverse strand.
			if (fillRead(reverseRead, tOccurThresh2)) {
				done = mergeAndVerifyK(true, k, reverseRead, reverseCandidates, reverseResults);
			}
		}
		// If still k > 0 run for remaining dist threshold on both forward and reverse strands.
		if (!done) {
			done = mergeAndVerifyK(false, k, forwardRead, forwardCandidates, forwardResults);
		}
		if (!done) {
			done = mergeAndVerifyK(false, k, reverseRead, reverseCandidates, reverseResults);
		}
	}
}

void SingleEndSearcher::mergeAndVerifyAll(Read& read, vector<unsigned>& candidates,
		vector<ReadMapping>& results) {

	if(indexer->useBitVector){
		if(read.additionalPrefix == 0) merger->findCandidatesAll(read, candidates);
		else merger->findCandidatesAll(read.bitvectorInvertedLists, read.prefixGrams, candidates);
	}
	else{
		if(read.additionalPrefix == 0) 
			merger->findCandidatesAllNF(read.invertedLists, read.prefixGrams, candidates);
		else
			merger->findCandidatesAll(read.invertedLists, read.prefixGrams, candidates);
	}

	verifier->verifyCandidates(*read.readStr, candidates, results, conf.uniqueMapping);
}

bool SingleEndSearcher::mergeAndVerifyK(bool exactMatch, unsigned& k, Read& read,
		vector<unsigned>& candidates, vector<ReadMapping>& results) {
	// Find exact matches first.
	if (exactMatch) {
		return (merger->findCandidatesAndExactMappingsK(k, read, candidates, results, *verifier));
	} else {
		return (merger->findCandidatesAndMappingsK(k, read, candidates, results, *verifier));
	}
}

void SingleEndSearcher::genReverses(ReadInfo& inputReadStructure) {
	// Generating reverse complement of a read.
	inputReadStructure.setReverseReadString(inputReadStructure.readString);
	getReverseComplementRead(inputReadStructure.reverseReadString);
	// Generating reverse read quality.
	inputReadStructure.setReverseReadQuality(inputReadStructure.readQuality);
	getReverseReadQuality(inputReadStructure.reverseReadQuality);
}
